// system 
using System;

// Bricsys
using Teigha.Runtime;
using Teigha.DatabaseServices;
using Teigha.Geometry;

using Bricscad.ApplicationServices;
using Bricscad.EditorInput;


namespace SolidSection
{
    public class Commands
    {
        [CommandMethod("SS")]
        public void SectionSolid()
        {
            Document doc = Application.DocumentManager.MdiActiveDocument;
            Database db = doc.Database;
            Editor ed = doc.Editor;

            // Ask the user to select an entity to section

            PromptEntityOptions peo = new PromptEntityOptions("\nSelect entity to section: ");

            peo.SetRejectMessage("\nEntity must be a 3D solid, " + "surface, body or region.");
            peo.AddAllowedClass(typeof(Solid3d), true);
            peo.AddAllowedClass(typeof(Teigha.DatabaseServices.Surface), true);
            peo.AddAllowedClass(typeof(Body), true);
            peo.AddAllowedClass(typeof(Region), true);

            PromptEntityResult per = ed.GetEntity(peo);

            if (per.Status != PromptStatus.OK)
                return;

            ObjectId entId = per.ObjectId;

            // Ask the user to define a section plane

            Point3dCollection pts = new Point3dCollection();

            PromptPointResult ppr = ed.GetPoint("\nPick first point for section: ");

            if (ppr.Status != PromptStatus.OK)
                return;

            pts.Add(ppr.Value);

            PromptPointOptions ppo = new PromptPointOptions("\nPick end point for section: ");
            ppo.BasePoint = ppr.Value;
            ppo.UseBasePoint = true;

            ppr = ed.GetPoint(ppo);

            if (ppr.Status != PromptStatus.OK)
                return;

            pts.Add(ppr.Value);

            // Ask what type of section to create


            PromptKeywordOptions pko = new PromptKeywordOptions("Enter section type ");
            pko.AllowNone = true;
            pko.Keywords.Add("2D");
            pko.Keywords.Add("3D");
            pko.Keywords.Add("Live");
            pko.Keywords.Default = "3D";

            PromptResult pkr = ed.GetKeywords(pko);

            if (pkr.Status != PromptStatus.OK)
                return;

            SectionType st;
            if (pkr.StringResult == "2D")
                st = SectionType.Section2d;
            else if (pkr.StringResult == "Live")
                st = SectionType.LiveSection;
            else // pkr.StringResult == "3D"
                st = SectionType.Section3d;

            // Now we're ready to do the real work

            Transaction tr = db.TransactionManager.StartTransaction();
            using (tr)
            {
                try
                {
                    BlockTable bt = (BlockTable)tr.GetObject(db.BlockTableId, OpenMode.ForRead);
                    BlockTableRecord ms = (BlockTableRecord)tr.GetObject(bt[BlockTableRecord.ModelSpace], OpenMode.ForWrite);

                    // Now let's create our section

                    Section sec = new Section(pts, Vector3d.ZAxis);
                    sec.State = SectionState.Plane;
            
                    // The section must be added to the drawing

                    ObjectId secId = ms.AppendEntity(sec);
                    tr.AddNewlyCreatedDBObject(sec, true);

                    // Set up some of its direct properties

                    sec.SetHeight(SectionHeight.HeightAboveSectionLine, 3.0);
                    sec.SetHeight(SectionHeight.HeightBelowSectionLine, 1.0);


                    // !!!!!!!!!!! clone the section !!!!!!!!!!!!!!!!
                    Section secCl = (Section)sec.Clone();



                    // ... and then clone its settings

                    SectionSettings ss = (SectionSettings)tr.GetObject(sec.Settings, OpenMode.ForWrite);

                    // !!!!!!!!  settings of the cloned section 
                    SectionSettings ssCl = (SectionSettings)tr.GetObject(secCl.Settings, OpenMode.ForWrite);

                    // Set our section type

                    ss.CurrentSectionType = st;

                    // !!!!!!!!!! type of clone section 
                    ssCl.CurrentSectionType = st;

                    // We only set one additional option if "Live"

                    if (st == SectionType.LiveSection)

                        //sec.EnableLiveSection(true);

                        // for ACAD 2010 >>>>
                        sec.IsLiveSectionEnabled = true;
                    else
                    {
                        // Non-live (i.e. 2D or 3D) settings

                        ObjectIdCollection oic = new ObjectIdCollection();
                        oic.Add(entId);

                        ss.SetSourceObjects(st, oic);

                        // !!!!!!!!!!!!!! SourceObjects for the clone section..
                        ssCl.SetSourceObjects(st, oic);

                        if (st == SectionType.Section2d)
                        {
                            // 2D-specific settings

                            //ss.SetVisibility(SectionType.Section2d, SectionGeometry.IntersectionFill, true);
                            //ss.SetVisibility(SectionType.Section2d, SectionGeometry.CurveTangencyLines, false);
                            //ss.SetVisibility(SectionType.Section2d, SectionGeometry.CurveTangencyLines, true);
                            //ss.SetVisibility(SectionType.Section2d, SectionGeometry.BackgroundGeometry, true);
                            //ss.SetHiddenLine(SectionType.Section2d, SectionGeometry.BackgroundGeometry, true);

                            ss.SetVisibility(st, SectionGeometry.BackgroundGeometry, true);
                            //generate background geometry with hidden entities !
                            ss.SetHiddenLine(st, SectionGeometry.BackgroundGeometry, true);

                            ssCl.SetVisibility(st, SectionGeometry.BackgroundGeometry, true);
                            // !!!!!!!!!  generate background geometry without hidden entities in the cloned section !
                            ssCl.SetHiddenLine(st, SectionGeometry.BackgroundGeometry, false);

                            Teigha.Colors.Color c = new Teigha.Colors.Color();
                            c = Teigha.Colors.Color.FromColorIndex(Teigha.Colors.ColorMethod.ByLayer, 22);

                            ss.SetColor(SectionType.Section2d, SectionGeometry.BackgroundGeometry, c);
                            ss.SetLayer(SectionType.Section2d, SectionGeometry.BackgroundGeometry, "BackgroundGeometryHidden");
                            ss.SetLayer(SectionType.Section2d, SectionGeometry.ForegroundGeometry, "ForegroundGeometryHidden");
                            ss.SetLayer(SectionType.Section2d, SectionGeometry.IntersectionBoundary, "IntersectionBoundaryHidden");
                            ss.SetLayer(SectionType.Section2d, SectionGeometry.IntersectionFill, "IntersectionFillHidden");
                            ss.SetLayer(SectionType.Section2d, SectionGeometry.CurveTangencyLines, "CurveTangencyLinesHidden");
                            ss.SetLinetype(SectionType.Section2d, SectionGeometry.BackgroundGeometry, "DASHDOT");
                            ss.SetLinetypeScale(SectionType.Section2d, SectionGeometry.BackgroundGeometry, 0.2);
                            ss.SetLineWeight(SectionType.Section2d, SectionGeometry.BackgroundGeometry, LineWeight.ByLayer);

                            // for the cloned section...
                            c = Teigha.Colors.Color.FromColorIndex(Teigha.Colors.ColorMethod.ByLayer, 3);
                            ssCl.SetColor(SectionType.Section2d, SectionGeometry.BackgroundGeometry, c);
                            ssCl.SetLayer(SectionType.Section2d, SectionGeometry.BackgroundGeometry, "BackgroundGeometry");
                            ssCl.SetLineWeight(SectionType.Section2d, SectionGeometry.BackgroundGeometry, LineWeight.ByLayer);
                            ssCl.SetLayer(SectionType.Section2d, SectionGeometry.ForegroundGeometry, "ForegroundGeometry");
                            ssCl.SetLayer(SectionType.Section2d, SectionGeometry.IntersectionBoundary, "IntersectionBoundary");
                            ssCl.SetLayer(SectionType.Section2d, SectionGeometry.IntersectionFill, "IntersectionFill");
                            ssCl.SetLayer(SectionType.Section2d, SectionGeometry.CurveTangencyLines, "CurveTangencyLines");
                        }
                        else if (st == SectionType.Section3d)
                        {
                            // 3D-specific settings

                            ss.SetVisibility(st, SectionGeometry.ForegroundGeometry, true);
                        }

                        // Finish up the common 2D/3D settings

                        ss.SetGenerationOptions(
                          st,
                          SectionGeneration.SourceSelectedObjects |
                          SectionGeneration.DestinationFile
                        );

                        // !!!!!!!!!!! cloned section
                        ssCl.SetGenerationOptions(
                          st,
                          SectionGeneration.SourceSelectedObjects |
                          SectionGeneration.DestinationFile
                        );
                    }


                    // Add the geometry to the modelspace
                    // (start by combining the various arrays,
                    // so we then have one loop, not four)


                    // Open up the main entity

                    Entity ent = (Entity)tr.GetObject(entId, OpenMode.ForRead);

                    // Generate the section geometry

                    Array flEnts, bgEnts, fgEnts, ftEnts, ctEnts;
                    sec.GenerateSectionGeometry(
                      ent,
                      out flEnts,
                      out bgEnts,
                      out fgEnts,
                      out ftEnts,
                      out ctEnts
                    );
                    
                    // !!!!!!!!!  Generate the cloned section geometry without hidden lines

                    Array flEntsCl, bgEntsCl, fgEntsCl, ftEntsCl, ctEntsCl;
                    secCl.GenerateSectionGeometry(
                      ent,
                      out flEntsCl,
                      out bgEntsCl,
                      out fgEntsCl,
                      out ftEntsCl,
                      out ctEntsCl
                    );
                   

                    int numEnts =
                        flEnts.Length + fgEnts.Length +
                        bgEnts.Length + ftEnts.Length +
                        ctEnts.Length;

                    // Create the appropriately-sized array

                    Array ents =
                    Array.CreateInstance(
                      typeof(Entity),
                      numEnts
                    );


                    // Copy across the contents of the
                    // various arrays

                    int index = 0;
                    flEnts.CopyTo(ents, index);
                    index += flEnts.Length;
                    fgEnts.CopyTo(ents, index);
                    index += fgEnts.Length;
                    bgEnts.CopyTo(ents, index);
                    index += bgEnts.Length;
                    ftEnts.CopyTo(ents, index);
                    index += ftEnts.Length;
                    ctEnts.CopyTo(ents, index);

                    // Our single loop to add entities
                    foreach (Entity ent2 in ents)
                    {
                        ms.AppendEntity(ent2);
                        tr.AddNewlyCreatedDBObject(ent2, true);
                    }

                    // add the entities of the cloned section without hidden lines for the background geometry...!

                    foreach (Entity backgrEnt in bgEntsCl)
                    {
                      ms.AppendEntity(backgrEnt);
                      tr.AddNewlyCreatedDBObject(backgrEnt, true);
                    }
                    
                    foreach (Entity flEnt in flEntsCl)
                    {
                        ms.AppendEntity(flEnt);
                        tr.AddNewlyCreatedDBObject(flEnt, true);
                    }
                    
                    foreach (Entity fgEnt in fgEntsCl)
                    {
                        ms.AppendEntity(fgEnt);
                        tr.AddNewlyCreatedDBObject(fgEnt, true);
                    }
                    foreach (Entity ftEnt in ftEntsCl)
                    {
                        ms.AppendEntity(ftEnt);
                        tr.AddNewlyCreatedDBObject(ftEnt, true);
                    }
                    foreach (Entity ctEnt in ctEntsCl)
                    {
                        ms.AppendEntity(ctEnt);
                        tr.AddNewlyCreatedDBObject(ctEnt, true);
                    }

                    tr.Commit();
                }
                catch (System.Exception ex)
                {
                    ed.WriteMessage(
                      "\nException: " + ex.Message
                    );
                }
            }
        }
